# 24oct23 Software Lab. Alexander Burger

### Entity/Relations ###
#
#    nm                      nm   pw  name
#     |                       |   |   |
#  +--*--------+           +--*---*---*--+
#  |           |           |             |
#  |           | usr       |             *-- em
#  |   Role    @-----------*    User     |
#  |           |      role |             *-- jpg
#  |           |           |             |
#  +--------@--+           +--*---+---*--+
#           |                 |   |   |
#           perm             ca   |   new
#                                 |
#                                 |
#        nm                     u |   d   t                  nm
#        |                        |   |   |                  |
#     +--*------+              +--*---*---*--------+        +--*------+
#     |         | mup      doc |                   | bin    |         |
#     |         *--------------*                   @--------+   Bin   |
#     |   Doc   |              |        Mup        |        |         |
#     |         +--------------@                   *--+ >   +--*---*--+
#     |         |          rpl |                   |  |        |   |
#     +---------+              +--*---*---*-----*--+  |        d   mt
#                                 |   |   |     |     |
#                               txt  sum  min   +-----+
#                                               <

### Markup index ###
(de foldedWords (Mup)
   (when Mup
      (uniq
         (filter '((W) (>= (length W) 3))
            (splitWords (in (blob Mup 'txt) (till))) ) ) ) )

(class +MupIdx +index)

(dm has> (Val Mup)
   (if (isa '+Mup Val)
      (extra Val Mup)
      (and (member Val (foldedWords Mup)) Mup) ) )

(dm rel> (Obj Old New Hook)
   (let
      (Tree (tree (: var) (: cls) (or Hook (and (: hook) (get Obj @))))
         A (foldedWords Old)
         B (foldedWords New) )
      (for W (diff A B)
         (store Tree (cons W Obj) NIL (: dbf)) )
      (for W (diff B A)
         (store Tree (cons W Obj) Obj (: dbf)) ) )
   (extra Obj Old New Hook) )

(dm rel?> (Obj Val Hook)
   (let Tree (tree (: var) (: cls) (or Hook (and (: hook) (get Obj @))))
      (fully
         '((W) (fetch Tree (cons W Obj)))
         (foldedWords Val) ) ) )

(dm lose> (Obj Val Hook)
   (let Tree (tree (: var) (: cls) (or Hook (and (: hook) (get Obj @))))
      (for W (foldedWords Val)
         (store Tree (cons W Obj) NIL (: dbf)) ) )
   (extra Obj Val Hook) )

(dm keep> (Obj Val Hook)
   (let Tree (tree (: var) (: cls) (or Hook (and (: hook) (get Obj @))))
      (for W (foldedWords Val)
         (store Tree (cons W Obj) Obj (: dbf)) ) )
   (extra Obj Val Hook) )

(dm iter> (X Lst @Flg)
   (cons
      (init '(mup . +Doc) (cons X) (cons X T))
      step ) )

(dm match> (X Val Obj)
   (fetch '(mup . +Doc) (cons X Obj)) )


### User ###
(extend +User)
(put *Class 'em)
(undef 'url> *Class)

(rel name (+Sn +Idx +String))          # Full Name
(rel em (+Fold +Ref +String))          # EMail
(rel jpg (+Blob))                      # Picture
(rel ca (+Ref +Date))                  # Created
(rel new (+Ref +Bool))                 # New account

(dm T @
   (pass super 'ca (date) 'new T) )

(dm url> (Tab)
   (and (may Wiki) (list "wiki/user.l" '*ID This)) )


### Document ###
(class +Doc +Entity)
(rel nm (+Key +String))                # Document title
(rel mup (+MupIdx +Joint)              # Current markup
   NIL
   doc (+Mup) )

(dm url> (Tab)
   (and (may Wiki) (list "wiki/doc.l" '*ID This)) )

(dm newMup> (Txt Sum Min Bin)
   (unless
      (and
         (: mup)
         (= Txt (in (blob @ 'txt) (till NIL T)))
         (= Bin (: mup bin)) )
      (dbSync)
      (protect
         (let Mup
            (new (db: +Mup) '(+Mup)
               'u *Login
               'd (date)
               't (time T)
               '< (: mup)
               'sum Sum
               'min Min
               'rpl (: mup rpl)
               'bin Bin )
            (put> Mup 'txt T)
            (out (blob Mup 'txt)
               (prin Txt) )
            (blob+ Mup 'txt)
            (put> Mup 'doc This) )
         (commit 'upd) ) ) )


# Markup
(class +Mup +Entity)
(rel doc (+Joint) mup (+Doc))          # Document
(rel u (+Aux +Ref +Link) (d t)         # User
   NIL (+User) )
(rel d (+Aux +Ref +Date) (t))          # Date
(rel t (+Time))                        # Time
(rel < (+Joint) > (+Mup))              # Previous
(rel > (+Joint) < (+Mup))              # Next
(rel txt (+Blob))                      # Text
(rel sum (+String))                    # Summary
(rel min (+Bool))                      # Minor edit
(rel rpl (+List +Link) (+Doc))         # Replies
(rel bin (+List +Link) (+Bin))         # Binaries

(dm url> (Tab)
   (and (may Wiki) (list "wiki/mup.l" '*ID This)) )

(dm nm> ()
   (pack "<" (: u nm) " " (datSym (: d)) " " (tim$ (: t)) ">") )

(dm diff> ()
   (and (may Wiki) (list "wiki/diff.l" '*ID This)) )


# Binary data
(class +Bin +Entity)
(rel nm (+String))                     # Name
(rel mt (+String))                     # Mime type
(rel d (+Blob))                        # Binary data

(dm upd> (X Old)
   (when (== X 'd)
      (put> This 'mt (mimetype (blob This 'd))) ) )


### Database sizes ###
(dbs
   (3 +Role +User (+User pw))          # 512 Prevalent objects
   (0 +Doc +Bin)                       # A:64 Tiny objects
   (1 +Mup)                            # B:128 Small objects
   (2 (+Role nm))                      # C:256 Small indexes
   (4 (+User nm name em ca new))       # D:1024 Normal indexes
   (4 (+Doc nm))                       # E:1024
   (4 (+Mup u))                        # F:1024
   (4 (+Mup d))                        # G:1024
   (6 (+Doc mup)) )                    # H:4096 Full-text index
